#include "libirsig.h"
#include "libwait.h"

unsigned char irsig[IRSIGMAX];
static unsigned char sig_p;
static unsigned char sig_b;

/*------------------------------------------------------------------------------
 *
 *	赤外線信号バッファクリア
 *
 */
void clearIrSig() {

	for(sig_p = 0; sig_p < IRSIGMAX; sig_p++)	// clear signal buf
		irsig[sig_p] = 0;

	sig_b = -1; sig_p = 0;						// 最初のビットを捨てる
}

//	Example: AEHA IR signal format
// 
//		A		B		C				A		B						C
//	----+		+-------+	H		----+		+-----------------------+	H
//		|		|		|				|		|						|
//		+-------+		+---L			+-------+						+---L
//	0:	  0.4ms	  0.4ms				1:	  0.4ms			  1.2ms

/*------------------------------------------------------------------------------
 *
 *	赤外線信号受信処理
 *
 *	clearIrSig();
 *	while(IRSNS);								// wait Falling edge of IRSNS (A)
 *	receiveIrSig();
 */
void receiveIrSig() {

	unsigned char sig_span;

	while(1) {
		while(!IRSNS);							// wait Raising edge of IRSNS (B)

		sig_span = 0;
		while(IRSNS) {							// wait Falling edge of IRSNS (C)
			waitS(2);							// 2 x 3 + 13 = 19us 消費
			if(++sig_span > 200) break;
		}
#if 0											// 1: for timing check
		irsig[sig_p] = sig_span;
		sig_p++;
#else
		irsig[sig_p] <<= 1;
		if(sig_span > 30)
			irsig[sig_p] |= 1;					// check Active (long span) signal

		if(++sig_b == 8) {
			sig_b = 0;
			sig_p++;
		}
#endif
		if(sig_span > 200 || sig_p == IRSIGMAX) break;
	}
}

/*------------------------------------------------------------------------------
 *
 *	赤外線信号送信処理
 *
 */
void transmitIrSig() {

	// Not implemented yet...
}

