#include <sdcc-lib.h>
#include <pic18fregs.h>	// /usr/share/sdcc/include/pic16/pic18f2420.h
#include "inline.h"
#include "libwait.h"
#include "libirsig.h"
#include "libbcd20.h"
#include "libeeprom.h"

static __code char __at 0x200000 id_location0 = 'A';
static __code char __at 0x200001 id_location1 = 'B';
static __code char __at 0x200002 id_location2 = 'C';
static __code char __at 0x200003 id_location3 = 'D';
static __code char __at 0x200004 id_location4 = 'E';
static __code char __at 0x200005 id_location5 = 'F';
static __code char __at 0x200006 id_location6 = 'G';
static __code char __at 0x200007 id_location7 = 'H';

static __code char __at (__CONFIG1H) configword1h =
	_IESO_OFF_1H &					// Oscillator Switchover mode disabled
	_FCMEN_OFF_1H &					// Fail-Safe Clock Monitor disabled
	_OSC_INTIO67_1H;				// Internal oscillator block, port function on RA6 and RA7

static __code char __at (__CONFIG2L) configword2l =
	_BORV_3_2L &					// Minimum setting
	_BOREN_OFF_2L &					// Brown-out Reset disabled in hardware and software
	_PWRT_OFF_2L;					// PWRT disabled

static __code char __at (__CONFIG2H) configword2h =
	_WDTPS_32768_2H &				// 1:32768
	_WDT_OFF_2H;					// WDT disabled (control is placed on the SWDTEN bit)

static __code char __at (__CONFIG3H) configword3h =
	_MCLRE_OFF_3H &					// RE3 input pin enabled; MCLR disabled
	_LPT1OSC_OFF_3H &				// Timer1 configured for higher power operation
	_PBADEN_ON_3H &					// PORTB<4:0> pins are configured as analog input channels on Reset
	_CCP2MX_PORTC_3H;				// CCP2 input/output is multiplexed with RC1

static __code char __at (__CONFIG4L) configword4l =
	_DEBUG_OFF_4L &					// Background debugger disabled, RB6 and RB7 configured as general purpose I/O pins
	_XINST_OFF_4L &					// Instruction set extension and Indexed Addressing mode disabled (Legacy mode)
	_LVP_OFF_4L &					// Single-Supply ICSP disabled
	_STVREN_ON_4L;					// Stack full/underflow will cause Reset

static __code char __at (__CONFIG5L) configword5l =
	_CP0_OFF_5L &					// Block 0 (000800-001FFFh) not code-protected
	_CP1_OFF_5L;					// Block 1 (002000-003FFFh) not code-protected

static __code char __at (__CONFIG5H) configword5h =
	_CPB_OFF_5H &					// Boot block (000000-0007FFh) not code-protected
	_CPD_OFF_5H;					// Data EEPROM not code-protected

static __code char __at (__CONFIG6L) configword6l =
	_WRT0_OFF_6L &					// Block 0 (000800-001FFFh) not write-protected
	_WRT1_OFF_6L;					// Block 1 (002000-003FFFh) not write-protected

static __code char __at (__CONFIG6H) configword6h =
	_WRTB_OFF_6H &					// Boot block (000000-0007FFh) not write-protected
	_WRTC_OFF_6H &					// Configuration registers (300000-3000FFh) not write-protected
	_WRTD_OFF_6H;					// Data EEPROM not write-protected

static __code char __at (__CONFIG7L) configword7l =
	_EBTR0_OFF_7L &					// Block 0 (000800-001FFFh) not protected from table reads executed in other blocks
	_EBTR1_OFF_7L;					// Block 1 (002000-003FFFh) not protected from table reads executed in other blocks

static __code char __at (__CONFIG7H) configword7h =
	_EBTRB_OFF_7H;					// Boot block (000000-0007FFh) not protected from table reads executed in other blocks

// データ記録バッファ(リングバッファ)
static __code char __at 0xF00000 eedata0[] = { 0x0D, 0x0E, 0x0A, 0x0D, 0x0B, 0x0E, 0x0E, 0x0F };	// 'dEAd', 'bEEF'
static __code char __at 0xF00080 eedata1[] = { 8, 0 };	// 書き込みポインタ, カウンタ記憶
#define	EERWPTR_ADR		0x80
#define	EECTPTR_ADR		0x81

#define	BUZ		PORTAbits.RA6

static unsigned char portc_buf = 0b00000000;

static unsigned char intr_cnt = 0;
static unsigned char intr_trg = 0;

void isr(void) __interrupt 1 {		// 1: HIGH priority, 2: LOW priority
	TMR0L = 131;					// 131 + 125 = 256
	intr_cnt++;
	intr_trg = 1;
	INTCONbits.TMR0IF = 0;			// Clear timer interrupt
}

/*			  _______
	NC		-|E3 U	7|-	IRSNS		7SEG:	SL-1063-50
	NC		-|A0	6|-	7SEGP		IRSNS:	SPS-448-1
	NC		-|1		5|-	7SEGP		TMPSNS:	MCP9700E/TO
	NC		-|2		4|-	7SEGP		SPK:	16ohm 0.1W
	Vref+	-|3		3|-	7SEGP
	NC		-|4		2|-	7SEGP
	TMPSNS	-|5		1|-	7SEGP
	GND		-|	   B0|-	7SEGP
	NC		-|7		 |-	Vcc
	SPK		-|A6	 |-	GND
	NC		-|C0	7|-	7SEG0(RX)
	NC		-|1		6|-	7SEG1(TX)
	NC		-|2		5|-	7SEG2
	7SEG.	-|3_____4|-	7SEG3
*/

void main(void) {

	enum modes {
		MODE_INIT,
		MODE_TIMER,
		MODE_COUNTER,
		MODE_TEMP,
		MODE_RECORD,
	} mode = MODE_INIT;				// 動作モード

	unsigned char timer = 65;		// タイマ値(残秒数)
	unsigned char counter = 0;		// カウンタ値
	unsigned char rec_ptr = 0xFC;	// 記録閲覧ポインタ

	unsigned char beep_tone = 0b00001000;	// ビープ音(高)
	unsigned char beep_len = 0;				// ビープ音(長)

	unsigned char seg7pats[] = {
	//	  -6543210
		0b10010000,					// 0		--1-
		0b11011011,					// 1/:	  3|	|2 o2
		0b10101000,					// 2/'		--4-
		0b11001000,					// 3	  6|	|5 o5
		0b11000011,					// 4		--0-
		0b11000100,					// 5/.
		0b10000100,					// 6
		0b11011001,					// 7
		0b10000000,					// 8
		0b11000000,					// 9
		0b10000001,					// A
		0b10000110,					// b
		0b10101110,					// c
		0b10001010,					// d
		0b10100100,					// E
		0b10100101,					// F/_
		0b10001110,					// o
	};
	unsigned char seg7bits[] = {	// 12:00
		0b10000000,					// | | L最下位
		0b01000000,					// | |
		0b00100000,					// | |
		0b00010000,					// L最上位
		0b00001000,					//   Lコロン
	};
	unsigned char seg7pos = 0;		// 32410
	unsigned char seg7vals[5] = {
		16, 12, 15, 12,  2,			// "cF'co"
	};

				//  76543210
	PORTA		= 0b00000000;
	LATA		= 0b00000000;
	ADCON1		= 0b00011010;		// Vref+ / 1010:AN4-AN0
	CMCON		= 0b00000111;
	TRISA		= 0b00101111;		// 1: 入力, 0: 出力

									//	RA0:AN0			RB2:AN8
									//	RA1:AN1			RB3:AN9
									//	RA2:AN2(Vref-)	RB1:AN10
									//	RA3:AN3(Vref+)	RB4:AN11
									//	RA5:AN4			RB0:AN12

	PORTB		= 0b00000000;
	LATB		= 0b00000000;
	TRISB		= 0b10000000;		// 1: 入力, 0: 出力

	PORTC		= portc_buf;
	LATC		= 0b00000000;
	TRISC		= 0b00000000;		// 1: 入力, 0: 出力
//	TRISC		= 0b11000000;		// 1: 入力, 0: 出力

	INTCON		= 0b11100000;		// 初期値:0b0000000x
									// 7:GIE/GIEH Global Interrupt Enable bit
									// 6:PEIE/GIEL Peripheral Interrupt Enable bit
									// 5 TMR0IE: TMR0 Overflow Interrupt Enable bit
									// 4:INT0IE INT0 External Interrupt Enable bit
									// 3:RBIE RB Port Change Interrupt Enable bit
									// 2:TMR0IF TMR0 Overflow Interrupt Flag bit
									// 1:INT0IF INT0 External Interrupt Flag bit
									// 0:RBIF RB Port Change Interrupt Flag bit

//	INTCON2		= 0b11110101;		// 初期値:0b11110101
									// 7:RBPU PORTB Pull-up Enable bit(1:Enable)
									// 6:INTEDG0 External Interrupt 0 Edge Select bit(1:Rising)
									// 5:INTEDG1 External Interrupt 1 Edge Select bit(1:Rising)
									// 4:INTEDG2 External Interrupt 2 Edge Select bit(1:Rising)
									// 2:TMR0IP TMR0 Overflow Interrupt Priority bit(1:High)
									// 0:RBIP RB Port Change Interrupt Priority bit(1:High)

//	INTCON3		= 0b11000000;		// 初期値:0b11000000
									// 7:INT2IP INT2 External Interrupt Priority bit(1:High)
									// 6:INT1IP INT1 External Interrupt Priority bit(1:High)
									// 4:INT2IE INT2 External Interrupt Enable bit(1:Enable)
									// 3:INT1IE INT1 External Interrupt Enable bit(1:Enable)
									// 1:INT2IF INT2 External Interrupt Flag bit
									// 0:INT1IF INT1 External Interrupt Flag bit

//	PIR1		= 0b00000000;		// 初期値:0b00000000(フラグ)
//	PIE1		= 0b00000000;		// 初期値:0b00000000(イネーブル)
//	IPR1		= 0b11111111;		// 初期値:0b11111111(プライオリティ)
									// 7:PSPIF Parallel Slave Port Read/Write Interrupt Flag bit
									// 6:ADIF A/D Converter Interrupt Flag bit
									// 5:RCIF EUSART Receive Interrupt Flag bit
									// 4:TXIF EUSART Transmit Interrupt Flag bit
									// 3:SSPIF Master Synchronous Serial Port Interrupt Flag bit
									// 2:CCP1IF CCP1 Interrupt Flag bit
									// 1:TMR2IF TMR2 to PR2 Match Interrupt Flag bit
									// 0:TMR1IF TMR1 Overflow Interrupt Flag bit

//	PIR2		= 0b00000000;		// 初期値:0b00000000(フラグ)
//	PIE2		= 0b00000000;		// 初期値:0b00000000(イネーブル)
//	IPR2		= 0b11111111;		// 初期値:0b11111111(プライオリティ)
									// 7:OSCFIF: Oscillator Fail Interrupt Flag bit
									// 6:CMIF: Comparator Interrupt Flag bit
									// 4:EEIF: Data EEPROM/Flash Write Operation Interrupt Flag bit
									// 3:BCLIF: Bus Collision Interrupt Flag bit
									// 2:HLVDIF: High/Low-Voltage Detect Interrupt Flag bit
									// 1:TMR3IF: TMR3 Overflow Interrupt Flag bit
									// 0:CCP2IF: CCP2 Interrupt Flag bit

	T0CON		= 0b11010101;		// 初期値:0b11111111
									// 7:TMR0ON Timer0 On/Off Control bit(1:Enable)
									// 6:T08BIT: Timer0 8-Bit/16-Bit Control bit(1:8-bit)
									// 5:T0CS: Timer0 Clock Source Select bit(0:Internal)
									// 4:T0SE: Timer0 Source Edge Select bit(1:High-to-low)
									// 3:PSA: Timer0 Prescaler Assignment bit(0:Assign)
									// 2~0:T0PS Timer0 Prescaler Select bits(000:1/2, 111:1/256)
									// 101:1/64, 4MHz -(1/4)-> 1MHz -(1/256)-> 3.9KHz -(1/256)-> 15.2Hz

//	T1CON		= 0b00000000;		// 初期値:0b00000000
									// 7:RD16 16-Bit Read/Write Mode Enable bit(1:Enable)
									// 5~4:T1CKPS Timer1 Input Clock Prescale Select bits(00:1/1, 11:1/8)
									// 3:T1OSCEN Timer1 Oscillator Enable bit(1:Enable)
									// 2:T1SYNC Timer1 External Clock Input Synchronization Select bit(0:Sync)
									// 1:TMR1CS Timer1 Clock Source Select bit(0:Internal)
									// 0:TMR1ON Timer1 On bit(1:Enable)

//	T2CON		= 0b00000000;		// 初期値:0b00000000
									// 6~3:T2OUTPS Timer2 Output Postscale Select bits(0000:1/1, 0011:1/4, 1111:1/16)
									// 2:TMR2ON Timer2 On bit(1:On)
									// 1~0:T2CKPS Timer2 Clock Prescale Select bits(00:1/1, 01:1/4, 1x:1/16)

	OSCCON		= 0b01100000;		// change to 4MHz
	while(!OSCCONbits.IOFS);

//	forCalibOsc();

	EEADR = EECTPTR_ADR;						// 記憶からカウンタを復帰
	readEeprom();
	counter = EEDATA;

	//--------------------------------------------------------------------------
	//
	//	Main Loop
	//
	while(1) {

		// 赤外線信号受信
		if(!IRSNS) {

			portc_buf &= 0b00000111;			// 7 セグを一時消灯
			PORTC = portc_buf;

			receiveIrSig();

			if(irsig[0] == 0xF1) {
				switch(irsig[1]) {

				  // タイマモードへ
				  case 0x21:					// VOL+ ボタン
					mode = MODE_TIMER;
					timer = 65;					// タイマリセット
					seg7vals[2] = 15;			// "cF:xx"(秒)
					seg7vals[3] = 12;
					beep_tone = 0b00001000;		// ビープ音を鳴らす
					beep_len = 8;
					break;

				  // カウンタモードへ
				  case 0xA1:					// VOL- ボタン
					if(mode == MODE_COUNTER)
						counter = 0;			// カウンタリセット
					mode = MODE_COUNTER;
					seg7vals[2] = 16;			// "co xx"(カウント)
					seg7vals[3] = 12;
					seg7vals[4] = 15;
					beep_tone = 0b00001000;		// ビープ音を鳴らす
					beep_len = 8;
					break;

				  // カウントアップ
				  case 0x29:					// U ボタン
					if(mode == MODE_COUNTER) ++counter;
					++timer;
					beep_tone = 0b00000100;		// 高いビープ音を鳴らす
					beep_len = 8;
					break;

				  // カウントダウン
				  case 0xA9:					// D ボタン
					if(mode == MODE_COUNTER) --counter;
					--timer;
					beep_tone = 0b00010000;		// 低いビープ音を鳴らす
					beep_len = 8;
					break;

				  // 温度計測
				  case 0xB1:					// SOUND ボタン
					mode = MODE_TEMP;
					seg7vals[0] = 12;			// "xx.xc"
					seg7vals[4] = 5;
					beep_tone = 0b00001000;		// ビープ音を鳴らす
					beep_len = 8;
					break;

				  // 情報記録
				  case 0x71:					// POWER ボタン
					EEADR = EERWPTR_ADR;		// 読み書きポインタ
					readEeprom();
					EEADR = EEDATA;				// 表示内容をそのまま記録
					EEDATA = seg7vals[3];
					writeEeprom();
					++EEADR;
					EEDATA = seg7vals[2];
					writeEeprom();
					++EEADR;
					EEDATA = seg7vals[1];
					writeEeprom();
					++EEADR;
					EEDATA = seg7vals[0];
					writeEeprom();
					++EEADR;
					EEDATA = EEADR & 0x7F;		// 読み書きポインタ更新
					EEADR = EERWPTR_ADR;
					writeEeprom();
					EEDATA = counter;			// カウンタを記憶
					EEADR = EECTPTR_ADR;
					writeEeprom();
					beep_tone = 0b00001000;		// 長いビープ音を鳴らす
					beep_len = 32;
					break;

				  // 情報閲覧モードへ
				  case 0x11:					// SOURCE ボタン
					mode = MODE_RECORD;
					rec_ptr = 0xFC;				// 記録閲覧ポインタ(初期値:-4(直前の記録))
					beep_tone = 0b00001000;		// ビープ音を鳴らす
					beep_len = 8;
					break;

				  // 情報閲覧、戻る
				  case 0xC9:					// R ボタン
					rec_ptr -= 4;
					if(mode == MODE_COUNTER) counter -= 10;
					timer -= 10;
					beep_tone = 0b00010000;		// 低いビープ音を鳴らす
					beep_len = 16;
					break;

				  // 情報閲覧、進む
				  case 0x49:					// F ボタン
					rec_ptr += 4;
					if(mode == MODE_COUNTER) counter += 10;
					timer += 10;
					beep_tone = 0b00000100;		// 高いビープ音を鳴らす
					beep_len = 16;
					break;
				}

				switch(mode) {

				  // タイマモード
				  case MODE_TIMER:
					b20bcd2 = 0;
					b20bcd1 = 0;
					b20bcd0 = timer;
					bin20_to_bcd();
					seg7vals[0] = b20bcd2 >> 4;
					seg7vals[1] = b20bcd3 & 0x0F;
					break;

				  // カウンタモード
				  case MODE_COUNTER:
					b20bcd2 = 0;
					b20bcd1 = 0;
					b20bcd0 = counter;
					bin20_to_bcd();
					seg7vals[0] = b20bcd2 >> 4;
					seg7vals[1] = b20bcd3 & 0x0F;
					break;

				  // 情報閲覧モード
				  case MODE_RECORD:
					EEADR = EERWPTR_ADR;		// 読み書きポインタ
					readEeprom();
					EEADR = (EEDATA + rec_ptr) & 0x7F;
					readEeprom();
					seg7vals[3] = EEDATA;
					++EEADR;
					readEeprom();
					seg7vals[2] = EEDATA;
					++EEADR;
					readEeprom();
					seg7vals[1] = EEDATA;
					++EEADR;
					readEeprom();
					seg7vals[0] = EEDATA;
					break;
				}
			}

			clearIrSig();
		}

		// 1 秒間隔の処理
		if(intr_cnt > 124) {
			intr_cnt = 0;

			// 減算タイマ
			if(mode == MODE_TIMER) {
				if(--timer != 0xFF) {
					b20bcd2 = 0;
					b20bcd1 = 0;
					b20bcd0 = timer;
					bin20_to_bcd();
					seg7vals[0] = b20bcd2 >> 4;
					seg7vals[1] = b20bcd3 & 0x0F;
				} else {
					seg7vals[0] = 16;			// "cf do"
					seg7vals[1] = 13;
					seg7vals[4] = 15;
					beep_tone = 0b00001000;		// 低いビープ音を鳴らす
					beep_len = 16;
					mode = MODE_INIT;
				}
			}

			// 温度計測
			if(mode == MODE_TEMP) {
				portc_buf &= 0b00000111;		// 7 セグを一時消灯
				PORTC = portc_buf;

				ADCON0 = 0b00010001;			// AN4 / ADON
				ADCON2 = 0b10000100;			// Right / 0Tad / Fosc/4
				waitS(1);						// 1 x 3 + 13 = 16us 待つ
				ADCON0bits.GO = 1;
				while(ADCON0bits.NOT_DONE);		// Temp Sensor: MCP9700E/TO
												// 0.5V + c * 0.01V, 0.758V -> 25.8c, 0.999V -> 49.9c
												// 1.024V / 1024 = 0.001V -> 0.1c
				b20bcd2 = 0;
				b20bcd1 = ADRESH;
				b20bcd0 = ADRESL;
				bin20_to_bcd();
				seg7vals[1] = b20bcd2 >> 4;
				seg7vals[2] = b20bcd3 & 0x0F;
				seg7vals[3] = b20bcd3 >> 4;
			}
		}

		// 1/125 秒(8ms)間隔の処理
		if(intr_trg) {							// 1/125 秒(8ms)間隔の処理
			intr_trg = 0;

			// コロンの点滅
			if(mode == MODE_TIMER)
				seg7vals[4] = (intr_cnt < 90) ? 1 : 15;

			// 7 セグのダイナミック点灯
			if(++seg7pos > 4) seg7pos = 0;
			portc_buf &= 0b00000111;
			portc_buf |= seg7bits[seg7pos];
			PORTC = portc_buf;
			PORTB = seg7pats[seg7vals[seg7pos]];

			// ビープ音長管理
			if(beep_len > 0)
				beep_len--;
		}

		// ビープ音を鳴らす
		if(beep_len > 0) {
			if(TMR0L & beep_tone)	 {			// TMR0L: (131 -> 255) / 8ms
				BUZ	= 1;						// 10000011 -> 11111111
			} else {							//  ^8ms    ->  125Hz
				BUZ	= 0;						//     ^1ms -> 1000Hz
			}
		}
	}
}

// JVC RM-RK50 IR signals
// 00000000 = F1 71 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '.q..............	// POWER
// 00000020 = F1 B1 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '................	// SOUND
// 00000040 = F1 29 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '.)..............	// U
// 00000060 = F1 C9 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '................	// R
// 00000080 = F1 11 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '................	// SOURCE
// 000000a0 = F1 49 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '.I..............	// F
// 000000c0 = F1 A9 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '................	// D
// 000000e0 = F1 A1 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '................	// VOL-
// 00000100 = F1 21 01 00 00 00 00 00 00 00 00 00 00 00 00 FF '.!..............	// VOL+

